package ca.ubc.cs.mylyn.monitor.analysis;

import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.sql.Date;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang.time.DurationFormatUtils;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.text.source.ILineRange;
import org.eclipse.mylyn.internal.monitor.usage.InteractionEventLogger;
import org.eclipse.mylyn.monitor.core.InteractionEvent;

import ca.ubc.cs.mylyn.monitors.parts.AbstractTextEditorMonitor;
import ca.ubc.cs.mylyn.monitors.parts.analysis.ViewportPayload;

/**
 * A small framework for processing viewport update events, keeping track of
 * what elements are visible, and providing notifications of elements appearing
 * and disappearing (see {@link #elementAppeared(String, long)} and
 * {@link #elementDisappeared(String, long, long)}).  Subclasses can override
 * {@link #updateViewport(long, InteractionEvent, ViewportPayload, long)}
 * to be notified of when a viewport has finished being processed. 
 * @author Brian de Alwis
 *
 */
public abstract class AbstractViewportTrackingAnalyzer {

	/**
	 * Time (in milliseconds since epoch) of the first event processed.
	 */
	protected long firstTime;

	/**
	 * Map of IJavaElement handleIdentifiers to the time offset when added
	 */
	protected Map<String,Long> currentlyVisible;
	protected Map<String,ILineRange> lineMapping;
	protected Set<String> relevantElements = null;
	protected String currentVPStructureHandle = "";
	
	protected boolean verbose = false;

	public AbstractViewportTrackingAnalyzer() {}

	/**
	 * Notification that the element described by <CODE>handle</CODE>
	 * has disappeared from the viewport, having been visible for
	 * <code>milliseconds</code> milliseconds.
	 * @param handle describing the element previously visible
	 * @param duration the duration of its visibility in milliseconds
	 * @param timeOffset time in milliseconds since the start of this trace
	 */
	protected abstract void elementDisappeared(String handle, long duration, long timeOffset);
	
	/**
	 * Notification that the element described by <code>handle</code>
	 * has just appeared on screen.
	 * @param handle the element handle identifier
	 * @param timeOffset time in milliseconds since the start of this trace
	 */
	protected abstract void elementAppeared(String handle, long timeOffset);
	
	/** UTILITY FUNCTIONS */
	/**
	 * Return true if <code>elr</code> overlaps in any way with <code>vlr</code>.
	 * @param elr the containing range
	 * @param vlr the contained range
	 * @return true if <code>vlr</code> is fully- or partially-contained within <code>elr</code>
	 */
	public static boolean isOverlapping(ILineRange elr, ILineRange vlr) {
		return contains(elr, vlr.getStartLine()) 
		|| contains(elr, vlr.getStartLine() + vlr.getNumberOfLines())
		|| contains(vlr, elr.getStartLine());
	}

	protected static boolean contains(ILineRange elr, int lineNumber) {
		return lineNumber >= elr.getStartLine()
				&& lineNumber <= elr.getStartLine() + elr.getNumberOfLines();
	}
	
	/** AVAILABLE TO SUBCLASSES */
	public Collection<String> getVisibleHandles() {
		return currentlyVisible.keySet();
	}
	
	/** IMPLEMENTATION */

	/**
	 * Process the given list of events.  Assumes that the events list
	 * has been sorted by the event start dates.
	 */
	public void process(List<InteractionEvent> events) {
		InteractionEvent last = events.get(0); 
		firstTime = last.getDate().getTime();
		resetVisible(0);
		for(int i = 0; i < events.size(); i++) {
			InteractionEvent current = events.get(i);
			long currentTimeOffset = current.getDate().getTime() - firstTime;
			processEvent(current, currentTimeOffset, 
					current.getDate().getTime() - last.getDate().getTime());
		}
		verbify(events.get(events.size() - 1).getDate().getTime() - firstTime,
				"resetVisible since last event");
		resetVisible(events.get(events.size() - 1).getDate().getTime() - firstTime);
	}
	
	/**
	 * Process the event 
	 * @param current the event being considered
	 * @param timeOffset
	 * @param eventDelta milliseconds since the previous event
	 */
	protected void processEvent(InteractionEvent current, long timeOffset, long eventDelta) {
		if(isTooLongToCount(eventDelta)) {
			System.err.println("Warning: time difference between events > 4mins; ignoring: "
				+ DurationFormatUtils.formatDurationHMS(eventDelta));
			resetVisible(timeOffset);
		}
		ViewportPayload vp = loadViewportPayload(current);
		if(vp != null) {
			updateViewport(timeOffset, current, vp, eventDelta);
		}
	}

	protected ViewportPayload loadViewportPayload(InteractionEvent event) {
		if ((event.getDelta().startsWith("<")) 
				&& ((JavaUI.ID_CF_EDITOR.equals(event.getOriginId())) 
						|| (JavaUI.ID_CU_EDITOR.equals(event.getOriginId())))
				&& !AbstractTextEditorMonitor.OPENED.equals(event.getNavigation())) {
			// OPENED is sent the first time a file is opened (i.e., a new editor)
			// But it reports the lines that *would* be visible if the entirety
			// could be displayed. Always followed by a BROUGHT_TO_TOP
			// with what is *actually* displayed.
			try {
				return ViewportPayload.load(event.getDelta());
			} catch (RuntimeException e) {
				System.err.println("Exception while parsing viewport payload: "
						+ e.getMessage());
				System.err.println("   event at: " + event.getDate());
			}
		}
		return null;
	}

	/**
	 * Determine whether a delta (in milliseconds) between viewport changes
	 * is deemed to be too long.  If true, then it's expected that the user left 
	 * for a break, or something, and returns having to refresh what they
	 * were looking at. This essentially wipes the slate clean, so that what is visible is
	 * treated as if it had just been seen.  Note: the event delta is not
	 * equivalent to the time between <em>events</em>, but between 
	 * <em>viewport changes</em>; that is, changes in the editor content.
	 * @param vpChangeDelta milliseconds since last viewport change
	 * @return true if this is considered too long
	 */
	protected boolean isTooLongToCount(long vpChangeDelta) {
		return false;	// available to override
		// return eventDelta > 4 * 60 * 1000;
	}

	/**
	 * There has been some change in the viewport such that
	 * all previously-visible elements are no longer visible.
	 * @param currentTime the time at which this change occured.
	 */
	protected void resetVisible(long currentTimeOffset) {
		if(currentlyVisible != null) {
			for(String identifier : currentlyVisible.keySet()) {
				elementDisappeared(identifier,
						currentTimeOffset - currentlyVisible.get(identifier),
						currentTimeOffset);
			}
		}
		currentlyVisible = new HashMap<String,Long>();
		lineMapping = new HashMap<String,ILineRange>();
		currentVPStructureHandle = "";
	}

	protected void updateViewport(long currentTimeOffset, InteractionEvent event, ViewportPayload vp, long eventDelta) {
		if(vp.getType().equals(AbstractTextEditorMonitor.OPENED)
				|| vp.getType().equals(AbstractTextEditorMonitor.BROUGHT_TO_TOP)) {
			verbify(currentTimeOffset, "vp reset (" + vp.getType() + "): " +
					event.getStructureKind() + ":" + event.getStructureHandle());
			resetVisible(currentTimeOffset);
		} else if(!currentVPStructureHandle.equals(event.getStructureHandle())) {
			verbify(currentTimeOffset, "in-editor content replacement: " +
				event.getStructureKind() + ":" + event.getStructureHandle());
			resetVisible(currentTimeOffset);
		}
		currentVPStructureHandle = event.getStructureHandle();

		// add any newly-visible elements
		for(String handle[] : vp.getHandles().keySet()) {
			if(!currentlyVisible.containsKey(handle[1])) {
				elementAppeared(handle[1], currentTimeOffset);
				currentlyVisible.put(handle[1], currentTimeOffset);
				lineMapping.put(handle[1], vp.getHandles().get(handle));
			}
		}
		// remove any no-longer-visible elements: go through the currently
		// visible line ranges, identify their corresponding handles, and then
		// remove any handles not identified as corresponding.
		// FIXME: there must surely be a more efficient way!
		Set<String> stillVisible = new HashSet<String>();
		for(ILineRange vlr : vp.getVisible()) {	// vlr = visible LR
			for(String handle : lineMapping.keySet()) {
				if(!stillVisible.contains(handle)) {
					ILineRange elr = lineMapping.get(handle);	// elr = element LR 
					if(isOverlapping(elr, vlr)) {
						stillVisible.add(handle);
					}
				}
			}
		}
		for(String handle : new ArrayList<String>(currentlyVisible.keySet())) {
			if(!stillVisible.contains(handle)) {
				long startTimeOffset = currentlyVisible.remove(handle);
				lineMapping.remove(handle);
				long msVisible = currentTimeOffset - startTimeOffset;
				elementDisappeared(handle, msVisible, currentTimeOffset);
			}
		}
	}

	public static Comparator<InteractionEvent> getEventComparator() {
		return new Comparator<InteractionEvent>() {
			public int compare(InteractionEvent o1, InteractionEvent o2) {
				return o1.getDate().compareTo(o2.getDate());
			}};
	}

	public static List<InteractionEvent> loadEventsFromFile(File file) {
		InteractionEventLogger log;
		try {
			log = new InteractionEventLogger(File.createTempFile("xplor", null));
			List<InteractionEvent> events = log.getHistoryFromFile(file);
			Collections.sort(events, AbstractViewportTrackingAnalyzer.getEventComparator());
			return events;
		} catch (IOException e) {
			e.printStackTrace();		// shouldn't happen
			return null;
		}
	}

	protected static void writeAsMinutes(Writer writer, long milliseconds) throws IOException {
		writer.append(Float.toString((float)milliseconds / 1000 / 60));
	}

	public boolean isVerbose() {
		return verbose;
	}

	public void setVerbose(boolean verbose) {
		this.verbose = verbose;
	}
	
	public void verbify(Date d, String text) {
		verbify(d.getTime() - firstTime, text);
	}
	
	public void verbify(long timeOffset, String text) {
		if(!verbose) { return; }
		
		System.out.println(Long.toString(timeOffset / 1000 / 60)
				+ ':' + Long.toString(timeOffset / 1000 % 60) 
				+ '.' + Long.toString(timeOffset % 1000)
				+ ": " + text);
	}

	public void verbify(long startOffset, long endOffset, String text) {
		if(!verbose) { return; }
		
		System.out.println(Long.toString(startOffset / 1000 / 60)
				+ ':' + Long.toString(startOffset / 1000 % 60) 
				+ '.' + Long.toString(startOffset % 1000)
				+ '-' + Long.toString(endOffset / 1000 / 60)
				+ ':' + Long.toString(endOffset / 1000 % 60) 
				+ '.' + Long.toString(endOffset % 1000)
				+ ": " + text);
	}
}
