package ca.ubc.cs.mylyn.monitor.usage.editor;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.viewers.ColumnLayoutData;
import org.eclipse.jface.viewers.ColumnPixelData;
import org.eclipse.jface.viewers.ColumnWeightData;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TableLayout;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.window.Window;
import org.eclipse.mylyn.context.core.ContextCorePlugin;
import org.eclipse.mylyn.internal.context.core.InteractionContext;
import org.eclipse.mylyn.internal.context.core.InteractionContextExternalizer;
import org.eclipse.mylyn.internal.context.core.InteractionContextManager;
import org.eclipse.mylyn.internal.monitor.usage.InteractionEventLogger;
import org.eclipse.mylyn.internal.monitor.usage.MonitorPreferenceConstants;
import org.eclipse.mylyn.internal.monitor.usage.UiUsageMonitorPlugin;
import org.eclipse.mylyn.monitor.core.InteractionEvent;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseTrackListener;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.IWorkbenchActionConstants;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.actions.ActionFactory;
import org.eclipse.ui.dialogs.ContainerGenerator;
import org.eclipse.ui.dialogs.SaveAsDialog;
import org.eclipse.ui.part.EditorPart;

public class MonitorHistoryEditor extends EditorPart
		implements IEditorPart, IResourceChangeListener, IModelListener<InteractionEvent> {
	protected ListModel<InteractionEvent> events = new ListModel<InteractionEvent>();
	protected boolean dirty = false;

	protected Composite control;
	protected Label summary;
	protected TableViewer viewer;
	protected InteractionEventLabelProvider labelProvider;
	protected String columnNames[];
	protected TableColumn columns[];
	protected ColumnLayoutData columnLayouts[];
	protected Action deleteAction;
    protected TextToolTip toolTip;
    
    protected enum SourceType { Unknown, InteractionEventLog, MylynContext; };
    protected SourceType sourceType = SourceType.Unknown;

	public MonitorHistoryEditor() {
		super();
	}

	@Override
	public void doSave(IProgressMonitor monitor) {
		IFile file = ((IFileEditorInput)getEditorInput()).getFile();
		saveTo(file, monitor);
		dirty = false;
		firePropertyChange(PROP_DIRTY);
	}

	@Override
	public void doSaveAs() {
		IPath newName = promptForSaveAsName();
		if(newName == null) { return; }
		try {
			ContainerGenerator cg = new ContainerGenerator(newName.removeLastSegments(1)); 
			IContainer parent = cg.generateContainer(new NullProgressMonitor());
			saveTo(parent.getFile(new Path(newName.lastSegment())), new NullProgressMonitor());
			firePropertyChange(PROP_DIRTY);
		} catch(CoreException e) {
			e.printStackTrace();
		}
	}
	
	protected void saveTo(IFile file, IProgressMonitor monitor) {
		File jFile = file.getLocation().toFile();
		jFile.delete();		// just to be sure
		switch(sourceType) {
		case InteractionEventLog:
			saveInteractionEventLog(jFile);
			break;
			
		case MylynContext:
			saveMylynContext(jFile);
			break;
			
		default:
			Activator.log(new Status(IStatus.ERROR, Activator.pluginId, "Invalid event source type: " + sourceType));
		return;
		}
		try {
			file.refreshLocal(0, monitor);
		} catch (CoreException e) {
			e.printStackTrace();
		}
	}

	private void saveMylynContext(File file) {
		InteractionContextExternalizer externalizer = new InteractionContextExternalizer();
		String identifier = file.getName();
		int index = identifier.indexOf(InteractionContextManager.CONTEXT_FILE_EXTENSION);
		if(index >= 0) { identifier = identifier.substring(0, index); }
		InteractionContext context = new InteractionContext(identifier, InteractionContextManager.getCommonContextScaling());
		for(InteractionEvent event : events.getElements()) {
			context.parseEvent(event);
		}
		externalizer.writeContextToXml(context, file);
	}

	private void saveInteractionEventLog(File file) {
		// Explicitly turn off obfuscation: the contents are already obfuscated, if they were
		// obfuscated in the first place.
		boolean obfuscating = UiUsageMonitorPlugin.getPrefs().getBoolean(
				MonitorPreferenceConstants.PREF_MONITORING_OBFUSCATE);
		UiUsageMonitorPlugin.getPrefs().setValue(
			MonitorPreferenceConstants.PREF_MONITORING_OBFUSCATE, false);

		InteractionEventLogger logger = new InteractionEventLogger(file);
		logger.startMonitoring();
		for(InteractionEvent event : events.getElements()) {
			logger.interactionObserved(event);
		}
		logger.stopMonitoring();
		UiUsageMonitorPlugin.getPrefs().setValue(
				MonitorPreferenceConstants.PREF_MONITORING_OBFUSCATE, obfuscating);
	}

	@Override
	public void init(IEditorSite site, IEditorInput input)
			throws PartInitException {
		/*
		 * ideally we'd accept IStorageEditorInput but the Mylyn InteractionEventLogger
		 * can only read from files 
		 */
		if(!(input instanceof IFileEditorInput)) {	
			throw new PartInitException("invalid input");
		}
		try {
			setSite(site);
			setInput(input);
			setPartName(input.getName());
			if(!((IFileEditorInput)input).getFile().getLocation().toFile().exists()) {
				throw new PartInitException("File does not actually exist!");
			}
			if(!refreshFromInput()) {
				throw new PartInitException("Unable to load from file");
			}
			ResourcesPlugin.getWorkspace().addResourceChangeListener(this,
					IResourceChangeEvent.POST_CHANGE | IResourceChangeEvent.PRE_DELETE);
		} catch(IOException e) {
			throw new PartInitException("Unable to open file: " + e.getMessage(), e);
		}
	}

	protected boolean refreshFromInput() throws IOException {
		IFileEditorInput fei = (IFileEditorInput)getEditorInput();
		File file = fei.getFile().getLocation().toFile();
		List<InteractionEvent> e;
		if(ContextCorePlugin.getContextManager().isValidContextFile(file)) {
			if((e  = loadMylynContext(file)) == null) {
				return false;
			}
			sourceType = SourceType.MylynContext;
		} else {
			if((e = loadInteractionEventLog(file)) == null) {
				return false;
			}
			sourceType = SourceType.InteractionEventLog;
		}
		Collections.sort(e, getEventComparator());
		
		final List<InteractionEvent> E = e;
		getEditorSite().getShell().getDisplay().asyncExec(new Runnable() {
			public void run() {
				events.setElements(E);	// this will cause a refresh()
				dirty = false;				
			}});
		return true;
	}

	private List<InteractionEvent> loadInteractionEventLog(File file) {
		try { 
			InteractionEventLogger log = 
				new InteractionEventLogger(File.createTempFile("xplor", null));
			return log.getHistoryFromFile(file);
		} catch(IOException e) {
			return null;
		}
	}

	private List<InteractionEvent> loadMylynContext(File file) {
		InteractionContextExternalizer externalizer = new InteractionContextExternalizer();
		InteractionContext context = externalizer.readContextFromXML("temp", file, InteractionContextManager.getCommonContextScaling());
		if(context == null) { return null; }
		return context.getInteractionHistory();
	}

	public static Comparator<InteractionEvent> getEventComparator() {
		return new Comparator<InteractionEvent>() {
			public int compare(InteractionEvent o1, InteractionEvent o2) {
				return o1.getDate().compareTo(o2.getDate());
			}};
	}

	@Override
	public boolean isDirty() {
		return dirty;
	}

	@Override
	public boolean isSaveAsAllowed() {
		return true;
	}

	@Override
	public void createPartControl(Composite parent) {
		control = new Composite(parent, SWT.NONE);
		control.setLayout(new GridLayout(1, false));
		summary = new Label(control, SWT.LEFT | /*SWT.WRAP | */ SWT.READ_ONLY);
		
		viewer = new TableViewer(control, SWT.VIRTUAL |
				SWT.BORDER | SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL);
		 viewer.getTable().setLayoutData(new GridData(GridData.FILL_BOTH));
//		viewer.setUseHashlookup(true);
		viewer.setContentProvider(new ModelListContentProvider());
		viewer.setLabelProvider(labelProvider = new InteractionEventLabelProvider());
		// viewer.setSorter(new ViewerSorter());
		// viewer.getTree().setLinesVisible(true);
		//viewer.getControl().setLayoutData(new GridData());
		viewer.getTable().setHeaderVisible(true);
		events.addListener(labelProvider);
		if(events.size() > 0) { labelProvider.setFirstEvent(events.get(0)); }
		events.addListener(this);

		getSite().setSelectionProvider(viewer);
		setupColumns();
		makeActions();
		hookContextMenu();
		configureToolTip();

		TableResizeAdapter resizer = new TableResizeAdapter(viewer.getTable());
		for(ColumnLayoutData cld : columnLayouts) {
			resizer.addColumnData(cld);
		}
		parent.addControlListener(resizer);
		viewer.setInput(events);
		viewer.setItemCount(events.size());
		updateSummary();
		dirty = false;
	}

	protected ISelectionProvider getSelectionProvider() {
		return getSite().getSelectionProvider();
	}
	
    protected ISelection getSelection() {
        return getSelectionProvider().getSelection();
    }

	protected void deleteSelection() {
		ISelection selection = getSelection();
		if(selection != null && selection instanceof IStructuredSelection) {
			IStructuredSelection s = (IStructuredSelection)selection;
			for(Object selected : s.toArray()) {
				if(selected != null && selected instanceof InteractionEvent) {
					modelDeleteItem((InteractionEvent)selected);
				}
			}
		}
		getSelectionProvider().setSelection(new StructuredSelection());
	}


	protected void modelDeleteItem(InteractionEvent event) {
		events.remove(event);		
	}


	protected void setupColumns() {
		columnNames = new String[IInteractionEventDisplayConstants.NUMBER_COLUMNS];
		columns = new TableColumn[IInteractionEventDisplayConstants.NUMBER_COLUMNS];
		columnLayouts = new ColumnLayoutData[IInteractionEventDisplayConstants.NUMBER_COLUMNS];
		
		columns[IInteractionEventDisplayConstants.COLUMN_START_TIME_DELTA] =
			new TableColumn(viewer.getTable(), SWT.LEFT);
		columnNames[IInteractionEventDisplayConstants.COLUMN_START_TIME_DELTA] = "Since";
		columnLayouts[IInteractionEventDisplayConstants.COLUMN_START_TIME_DELTA] =
			new ColumnPixelData(80);
		
		columns[IInteractionEventDisplayConstants.COLUMN_START_TIME] = 
			new TableColumn(viewer.getTable(), SWT.LEFT);
		columnNames[IInteractionEventDisplayConstants.COLUMN_START_TIME] =
			"Time";
		columnLayouts[IInteractionEventDisplayConstants.COLUMN_START_TIME] =
			new ColumnPixelData(70);
		
		columns[IInteractionEventDisplayConstants.COLUMN_DURATION] = 
			new TableColumn(viewer.getTable(), SWT.LEFT);
		columnNames[IInteractionEventDisplayConstants.COLUMN_DURATION] =
			"Duration";
		columnLayouts[IInteractionEventDisplayConstants.COLUMN_DURATION] =
			new ColumnPixelData(40);

		columns[IInteractionEventDisplayConstants.COLUMN_KIND] = 
			new TableColumn(viewer.getTable(), SWT.LEFT);
		columnNames[IInteractionEventDisplayConstants.COLUMN_KIND] =
			"Kind";
		columnLayouts[IInteractionEventDisplayConstants.COLUMN_KIND] =
			new ColumnPixelData(60);

		columns[IInteractionEventDisplayConstants.COLUMN_CHANGE_DELTA] = 
			new TableColumn(viewer.getTable(), SWT.LEFT);
		columnNames[IInteractionEventDisplayConstants.COLUMN_CHANGE_DELTA] =
			"Delta";
		columnLayouts[IInteractionEventDisplayConstants.COLUMN_CHANGE_DELTA] = 
			new ColumnWeightData(55,50);

		columns[IInteractionEventDisplayConstants.COLUMN_ORIGINID] = 
			new TableColumn(viewer.getTable(), SWT.LEFT);
		columnNames[IInteractionEventDisplayConstants.COLUMN_ORIGINID] =
			"OriginId";
		columnLayouts[IInteractionEventDisplayConstants.COLUMN_ORIGINID] =
			new ColumnWeightData(40,25);

		columns[IInteractionEventDisplayConstants.COLUMN_NAVIGATION] = 
			new TableColumn(viewer.getTable(), SWT.LEFT);
		columnNames[IInteractionEventDisplayConstants.COLUMN_NAVIGATION] =
			"Navigation";
		columnLayouts[IInteractionEventDisplayConstants.COLUMN_NAVIGATION] =
			new ColumnWeightData(20, 20);

		columns[IInteractionEventDisplayConstants.COLUMN_STRUCTURE_KIND] = 
			new TableColumn(viewer.getTable(), SWT.LEFT);
		columnNames[IInteractionEventDisplayConstants.COLUMN_STRUCTURE_KIND] =
			"Structure Kind";
		columnLayouts[IInteractionEventDisplayConstants.COLUMN_STRUCTURE_KIND] =
			new ColumnWeightData(25,20);

		columns[IInteractionEventDisplayConstants.COLUMN_STRUCTURE_HANDLE] = 
			new TableColumn(viewer.getTable(), SWT.LEFT);
		columnNames[IInteractionEventDisplayConstants.COLUMN_STRUCTURE_HANDLE] =
			"Handle";
		columnLayouts[IInteractionEventDisplayConstants.COLUMN_STRUCTURE_HANDLE] =
			new ColumnWeightData(55,50);

		columns[IInteractionEventDisplayConstants.COLUMN_INTEREST_CONTRIBUTION] = 
			new TableColumn(viewer.getTable(), SWT.RIGHT);
		columnNames[IInteractionEventDisplayConstants.COLUMN_INTEREST_CONTRIBUTION] =
			"Interest";
		columnLayouts[IInteractionEventDisplayConstants.COLUMN_INTEREST_CONTRIBUTION] =
			new ColumnPixelData(10);

		TableLayout tableLayout = new TableLayout();
		for(int i = 0; i < columns.length; i++) {
			tableLayout.addColumnData(columnLayouts[i]);
			columns[i].setText(columnNames[i]);
			columns[i].setResizable(true);
			columns[i].setWidth(columnLayouts[i] instanceof ColumnPixelData ?
					((ColumnPixelData)columnLayouts[i]).width :
						((ColumnWeightData)columnLayouts[i]).minimumWidth);
		}
		viewer.getTable().setLayout(tableLayout);
		viewer.setColumnProperties(columnNames);
	}

	@Override
	public void setFocus() {
		viewer.getControl().setFocus();
	}

    protected void configureToolTip() {
    	toolTip = new TextToolTip(viewer.getControl());
		viewer.getControl().addMouseTrackListener(new MouseTrackListener() {
			public void mouseEnter(MouseEvent e) {
				toolTip.hideToolTip();
			}

			public void mouseExit(MouseEvent e) {
				toolTip.hideToolTip();
			}

			public void mouseHover(MouseEvent e) {
				TableItem item = viewer.getTable().getItem(new Point(e.x, e.y));
				if(item != null && item.getData() != null
						&& item.getData() instanceof InteractionEvent) {
					InteractionEvent obj = (InteractionEvent)item.getData();
					//System.out.println(e.x + "," + e.y + ": " + obj);
					String text = getToolTipText(obj);
					//System.out.println("-----\nShowing text: " + text + "\n----\n");
					if(text != null && text.trim().length() > 0) {
						toolTip.showToolTip(text);
					} else {
						toolTip.hideToolTip();
					}
				} else {
					toolTip.hideToolTip();
				}
			}		
		});
    }

	protected String getToolTipText(InteractionEvent event) {
		StringBuilder buf = new StringBuilder();
		appendText(buf, event.getKind().toString(), "kind: ", "\n");
		appendText(buf, event.getOriginId(), "origin: ", "\n");
		appendText(buf, event.getNavigation(), "navig: ", "\n");
		appendText(buf, event.getStructureKind(), "struct: ", "");
		appendText(buf, event.getStructureHandle(), ": ", "\n");
		appendText(buf, event.getDelta(), "delta: ", "\n");
		return buf.substring(0, buf.length() - 1);	// trim final newline
	}

	protected void appendText(StringBuilder buf, String checkText, String pre, String post) {
		if(checkText == null || checkText.length() == 0 || checkText.equals("null")) { 
			return;
		}
		buf.append(pre);
		buf.append(checkText);
		buf.append(post);
	}
	
	public void makeActions() {
		deleteAction = new Action() {
			public void run() {
				deleteSelection();
			}
		};
		deleteAction.setText("Delete");
		deleteAction.setToolTipText("Deleted selected element");
		deleteAction.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().
			getImageDescriptor(ISharedImages.IMG_TOOL_DELETE));
//		deleteAction.setActionDefinitionId(ActionFactory.DELETE.getId());
		getEditorSite().getActionBars().setGlobalActionHandler(ActionFactory.DELETE.getId(),
				deleteAction);
	}
	
	protected void hookContextMenu() {
		MenuManager menuMgr = new MenuManager("#PopupMenu");
		menuMgr.setRemoveAllWhenShown(true);
		menuMgr.addMenuListener(new IMenuListener() {
			public void menuAboutToShow(IMenuManager manager) {
				MonitorHistoryEditor.this.fillContextMenu(manager);
			}
		});
		Menu menu = menuMgr.createContextMenu(viewer.getControl());
		viewer.getControl().setMenu(menu);
		getSite().registerContextMenu(menuMgr, viewer);
	}

	protected void fillContextMenu(IMenuManager manager) {
		if(toolTip != null) { toolTip.hideToolTip(); }

		manager.add(deleteAction);
		manager.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));
	}

	@Override
	public void dispose() {
		ResourcesPlugin.getWorkspace().removeResourceChangeListener(this);
		events.removeListener(labelProvider);
		super.dispose();
	}

	protected IPath promptForSaveAsName() {
		SaveAsDialog dialog = new SaveAsDialog(getSite().getShell());
//		dialog.setMessage("Save events to...");
		IFile current = ((IFileEditorInput)getEditorInput()).getFile();
		dialog.setOriginalFile(current);
		dialog.setBlockOnOpen(true);
		if(dialog.open() != Window.OK) {
			return null;
		}
		return dialog.getResult();
	}

	public void resourceChanged(IResourceChangeEvent event) {
		switch(event.getType()) {
		case IResourceChangeEvent.PRE_DELETE:
			// Do I need to do anything?
			break;
			
		case IResourceChangeEvent.POST_CHANGE:
			IResourceDelta delta = event.getDelta().findMember(((IFileEditorInput)getEditorInput()).getFile().getFullPath());
			if(delta != null) {
				if((delta.getFlags() & (IResourceDelta.CONTENT|IResourceDelta.REPLACED)) != 0) {
					try {
						// FIXME: if(dirty) { prompt user: file changed on disk: reload contents? }
						refreshFromInput();
					} catch (IOException e) {
						Activator.log(e);
					}
				}
			}
			break;
		}
		
	}


	protected void updateSummary() {
		StringBuffer buffer = new StringBuffer();
		if(events.size() > 0) {
			buffer.append("Start: ");
			buffer.append(events.get(0).getDate());
			buffer.append("  End: ");
			buffer.append(events.last().getEndDate());
			buffer.append("\nDuration: ");
			buffer.append(formatDurationHMSM(
					events.get(0).getDate().getTime(),
					events.last().getEndDate().getTime()));
			buffer.append("  Items: ");
			buffer.append(events.size());
		} else {
			buffer.append("0 events");
		}
		if(summary != null) {
			summary.setText(buffer.toString());
		}
	}

	public static String formatDurationHMSM(long start, long end) {
		StringBuffer buf = new StringBuffer();
		long diff = end - start;	// in milliseconds
		reverseFormatNumber((int)((end - start) % 1000), 3, buf);
		buf.append('.'); diff /= 1000;	// in seconds
		reverseFormatNumber((int)(diff % 60), 2, buf);
		buf.append(':'); diff /= 60;	// in minutes
		reverseFormatNumber((int)(diff % 60), 2, buf);
		buf.append(':'); diff /= 60;	// in hours
		reverseFormatNumber((int)diff, 2, buf);
		return buf.reverse().toString();
	}
	
	public static String formatDurationHMS(long start, long end) {
		StringBuffer buf = new StringBuffer();
		long diff = end - start;	// in milliseconds
		diff /= 1000;	// in seconds
		reverseFormatNumber((int)(diff % 60), 2, buf);
		buf.append(':'); diff /= 60;	// in minutes
		reverseFormatNumber((int)(diff % 60), 2, buf);
		buf.append(':'); diff /= 60;	// in hours
		reverseFormatNumber((int)diff, 2, buf);
		return buf.reverse().toString();
	}
	
	private static void reverseFormatNumber(int number, int magnitude, StringBuffer buffer) {
		for(int i = 0; i < magnitude; i++) {
			buffer.append(Character.forDigit(number % 10, 10));
			number /= 10;
		}
	}

	public void refresh() {
		getSite().getShell().getDisplay().asyncExec(new Runnable() {
			public void run() {
				if(!viewer.getControl().isDisposed()) {
					viewer.setItemCount(events.size());
					viewer.refresh();
					updateSummary();
				}
			}
		});
	}
	
	public void modelCleared(ListModel<InteractionEvent> model) {
		dirty = true;
		refresh();
		firePropertyChange(PROP_DIRTY);
	}

	public void modelElementsChanged(ListModel<InteractionEvent> model) {
		dirty = true;
		refresh();
		firePropertyChange(PROP_DIRTY);
	}

	public void elementAdded(ListModel<InteractionEvent> model, InteractionEvent event) {
		dirty = true;
		refresh();
		firePropertyChange(PROP_DIRTY);
	}

	public void elementRemoved(ListModel<InteractionEvent> model, InteractionEvent element) {
		dirty = true;
		refresh();
		firePropertyChange(PROP_DIRTY);
	}

}
