/*******************************************************************************
 * Copyright (c) 2004 - 2006 University Of British Columbia and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     University Of British Columbia - initial API and implementation
 *******************************************************************************/

package ca.ubc.cs.mylyn.monitors.parts;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IRegion;
import org.eclipse.jface.text.ITextViewerExtension5;
import org.eclipse.jface.text.IViewportListener;
import org.eclipse.jface.text.Region;
import org.eclipse.jface.text.source.IAnnotationModel;
import org.eclipse.jface.text.source.IAnnotationModelListener;
import org.eclipse.jface.text.source.ILineRange;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.text.source.ISourceViewerExtension2;
import org.eclipse.jface.text.source.LineRange;
import org.eclipse.jface.text.source.projection.ProjectionViewer;
import org.eclipse.mylyn.monitor.core.InteractionEvent;
import org.eclipse.mylyn.monitor.ui.MonitorUiPlugin;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.texteditor.AbstractTextEditor;

/**
 * A monitor for listening to changes in the visible content (or viewport) of
 *  text editors.  This class handles the details of determining the changes
 * (what lines were added, removed); subclasses may add more semantically
 * meaningful to the report.  Intended to be specialized for different types 
 * of editor.
 * @author Brian de Alwis
 * @see JavaEditorViewportMonitor
 */
public abstract class AbstractTextEditorMonitor
	extends AbstractWorkbenchPartMonitor
	implements IAnnotationModelListener {
	
	public static final String FOLDING = "folding";
	public static final String SCROLLING = "scrolling";
	public static final String OPENED = "opened";
	public static final String BROUGHT_TO_TOP = "brought-to-top";
	public static final String ATTR_REMOVED = "removed";
	public static final String ATTR_ADDED = "added";
	public static final String ATTR_VISIBLE = "visible";
	public static final String ATTR_KIND = "kind";

    protected static final List<ILineRange> EMPTY_LINERANGE_LIST = new ArrayList<ILineRange>(0);

	private Map<IWorkbenchPart,IViewportListener> viewportListeners =
		new WeakHashMap<IWorkbenchPart,IViewportListener>();

	/**
	 * Keep track of recent ranges for parts with SourceViewers. 
	 */
	private Map<IWorkbenchPart,List<ILineRange>> recentSVVisibleRegions =
		new WeakHashMap<IWorkbenchPart,List<ILineRange>>();

	private Map<IAnnotationModel,IEditorPart> modelToPart =
		new WeakHashMap<IAnnotationModel,IEditorPart>();

	/**
	 * Records for a test-and-set type operation as to whether a
	 * a viewport-change scrolling notification is pending.
	 */
	private Map<ISourceViewer,Object> pendingViewportChanges =
		new WeakHashMap<ISourceViewer,Object>();

// These methods were used in harmony with the PartDispatcher's scrolling
// mechanism.  They are no longer necessary with the discovery of the 
// IViewportListener mechanism, but we keep them here just in case.
//	public void preScroll(String partId, IWorkbenchPart part, Scrollable widget,
//			ScrollBar scrollBar) {
//		if(!testAndSetScrollNoticePending(scrollBar)) { return; }
//		if(widget.getVerticalBar() == scrollBar && part instanceof IEditorPart) {
//			ISourceViewer viewer = getSourceViewer((IEditorPart)part);
//			if(viewer != null) {
//				setPreScrollValue(scrollBar, determineVisibleLines(viewer));
//			}
//		}
//	}
//
//	public void postScroll(String partId, IWorkbenchPart part, Scrollable widget,
//			ScrollBar scrollBar) {
//		List<ILineRange> preScrollLineRange = resetScrollNoticePending(scrollBar);
//		List<ILineRange> postScrollLineRange = null;
//		if(widget.getVerticalBar() == scrollBar && part instanceof IEditorPart) {
//			postScrollLineRange = determineVisibleLines((IEditorPart)part);
//			if(postScrollLineRange != null) {
//				recentSVVisibleRegions.put(part, postScrollLineRange);
//			}
//		}
//		if(preScrollLineRange != null && postScrollLineRange != null && 
//				!equalRanges(preScrollLineRange, postScrollLineRange)) {
//			reportChange(SCROLLING, part, preScrollLineRange, postScrollLineRange);
//		}
//	}

	protected boolean equalRanges(List<ILineRange> r1, List<ILineRange> r2) {
		if(r1.size() != r2.size()) { return false; }
		for(int i = 0; i < r1.size(); i++) {
			ILineRange l1 = r1.get(i);
			ILineRange l2 = r2.get(i);
			if(l1.getStartLine() != l2.getStartLine() || l1.getNumberOfLines() != l2.getNumberOfLines()) {
				return false;
			}
		}
		return true;
	}


	protected void printRange(StringBuffer buffer, List<ILineRange> lr) {
		for(ILineRange l : lr) {
			printInterval(buffer, l);
			buffer.append(' ');
		}
	}
	
	protected void printInterval(StringBuffer buffer, ILineRange l) {
		buffer.append(l.getStartLine() + 1);
		if(l.getNumberOfLines() > 1) {
			buffer.append(':');
			buffer.append(l.getStartLine() + l.getNumberOfLines() - 1 + 1);
		}
	}

	/**
	 * Report a change. 
	 * @param changeType
	 * @param part
	 * @param previous
	 * @param current
	 */
	protected void reportChange(String changeType, IWorkbenchPart part,
			List<ILineRange> previous, List<ILineRange> current) {
		LineRangeDifferences d = new LineRangeDifferences(previous, current);		
		MonitorUiPlugin.getDefault().notifyInteractionObserved(
				createInteractionEvent(changeType, part,
						d.getAdded(), d.getRemoved(), current));
	}


	/**
	 * Create and return an InteractionEvent describing the change in the
	 * viewport as described by <em>added</em>, <em>removed</em>,
	 * and the currently visible lines in <em>current</em>.  Useful helper
	 * methods are printInterval() and printRange(). 
	 * @param changeType a string describing the reason for the change
	 * @param part the editor with the change
	 * @param added the set of lines added
	 * @param removed the set of lines removed
	 * @param current the currently visible set of lines
	 * @return an InteractionEvent
	 */
	protected InteractionEvent createInteractionEvent(String changeType,
			IWorkbenchPart part, LinkedList<ILineRange> added,
			LinkedList<ILineRange> removed, List<ILineRange> current) {
		String changeDelta = createChangeDelta(changeType,
				part, added, removed, current);
		return new InteractionEvent(InteractionEvent.Kind.SELECTION,
				getStructureKind(part), getStructureHandle(part),
					part.getSite().getId(), changeType, changeDelta, 0f);
	}
	
	protected abstract String getStructureKind(IWorkbenchPart part);
	protected abstract String getStructureHandle(IWorkbenchPart part);

	protected String createChangeDelta(String changeType, IWorkbenchPart part,
			List<ILineRange> added, List<ILineRange> removed, List<ILineRange> current) {
		StringBuffer buffer = new StringBuffer("<");
		buffer.append(changeType);
		buffer.append(" ");
		buffer.append(ATTR_ADDED);
		buffer.append("=\"");
		printRange(buffer, added);
		buffer.append("\" ");
		buffer.append(ATTR_REMOVED);
		buffer.append("=\"");
		printRange(buffer, removed);
		buffer.append("\" ");
		buffer.append(ATTR_VISIBLE);
		buffer.append("=\"");
		printRange(buffer, current);
		buffer.append('"');

		String handles = describeHandles(changeType, part, added, removed, current);
		if(handles.length() == 0) {
			buffer.append("/>");
		} else {
			buffer.append(">\n");
			buffer.append(handles);
			buffer.append("\n</");
			buffer.append(changeType);
			buffer.append('>');
		}
		return buffer.toString();
	}
	
	protected String describeHandles(String changeType, IWorkbenchPart part,
			List<ILineRange> added, List<ILineRange> removed, List<ILineRange> current) {
		return "";
	}
	
	/**
	 * Obtain the {@link ISourceViewer} corresponding to the provided
	 * <em>part</em>.  This method is intended to be overridden by subclasses.
	 * @param part the corresponding part
	 * @return <em>part</em>'s source viewer
	 */
	protected ISourceViewer getSourceViewer(IWorkbenchPart part) {
		if(part instanceof AbstractTextEditor) {
			try {
				return  (ISourceViewer)invokeMethod(part, "getSourceViewer", new Class[0], new Object[0]);
			} catch(Exception e) { /* do nothing */ }
		}
		return null;
	}


	protected List<ILineRange> determineVisibleLines(IWorkbenchPart part) {
		ISourceViewer viewer = getSourceViewer(part);
		if(viewer == null) { return null; }
		return determineVisibleLines(viewer);
	}

	protected List<ILineRange> determineVisibleLines(ISourceViewer viewer) {
 		List<ILineRange> visibleLines = new ArrayList<ILineRange>();
		// There's a bug sometimes tickled in the Eclipse code where a BadLocationException
		// is raised -- and as a result, almost all the lines of the file are reported.  See if this makes
 		// a differences
		if(!viewer.getTextWidget().isVisible()) {
			return visibleLines;	// empty
		}

		IDocument doc = viewer.getDocument();
		// TextViewer.getTopIndexStartOffset() is buggy in 3.1
		int visibleTopLine = viewer.getTopIndex();
		int visibleBottomLine = viewer.getBottomIndex();
		try {
			int visibleTopOffset = doc.getLineOffset(visibleTopLine);
			int visibleBottomOffset = doc.getLineOffset(visibleBottomLine);
			IRegion visibleRegion = new Region(visibleTopOffset, visibleBottomOffset - visibleTopOffset + doc.getLineLength(visibleBottomLine));
			int length = doc.getLength();
//			System.out.println("  doc chars[0:" + doc.getLength() + "] lines[0:" + doc.getNumberOfLines() + "]");
//			System.out.println("  viewer: viewablelines: " + visibleTopLine + ":" + visibleBottomLine);
//			System.out.println("  visRegion=" + visibleRegion.getOffset() + ":" + visibleRegion.getOffset() + visibleRegion.getLength());
			if(viewer instanceof ITextViewerExtension5) {
				ITextViewerExtension5 v = (ITextViewerExtension5)viewer;
//				System.out.print("    visible regions: " );
				for(IRegion r : v.getCoveredModelRanges(visibleRegion)) {
					int startLine = doc.getLineOfOffset(r.getOffset());
					int stopLine = r.getOffset() + r.getLength() >= length
							? doc.getNumberOfLines() 
							: doc.getLineOfOffset(r.getOffset() + r.getLength());
					visibleLines.add(new LineRange(startLine, stopLine - startLine));
//					System.out.print(startLine + ":" + stopLine + " ");
				}
//				System.out.println();
			} else {
				int startLine = doc.getLineOfOffset(visibleRegion.getOffset());
				int stopLine = visibleRegion.getOffset() + visibleRegion.getLength() >= length
					? doc.getNumberOfLines() 
							: doc.getLineOfOffset(visibleRegion.getOffset() + visibleRegion.getLength());
				visibleLines.add(new LineRange(startLine, stopLine - startLine));
//				System.out.println("  visible region: " + startLine + ":" + stopLine);
			}
		} catch(BadLocationException e) {
			System.err.println("Error calculating visible regions: " + e);
			visibleLines.add(new LineRange(visibleTopLine, visibleBottomLine - visibleTopLine));
		}
 		if(countVisibleLines(visibleLines) > viewer.getTextWidget().getLineCount()) {
			Activator.getDefault().getLog().log(new Status(IStatus.WARNING,
					Activator.PLUGIN_ID, -1, 
					getClass().getName() + ": invalid visible lines for "
					+ viewer + ": computed count (" + countVisibleLines(visibleLines)
					+ ") > actual from text widget (" + viewer.getTextWidget().getLineCount() +")",
					null));
			return EMPTY_LINERANGE_LIST;
		}
		return visibleLines;
	}

	protected List<ILineRange> updateSVVisibleRegions(IWorkbenchPart part) {
		List<ILineRange> list;
		if((list = determineVisibleLines(part)) != null) {
			recentSVVisibleRegions.put(part, list);
		}
		return list;
	}

	public void partOpened(IWorkbenchPart part) {
//		System.out.println("partOpened(" + part + ")");
		List<ILineRange> current = updateSVVisibleRegions(part);
		registerAnnotationModelListener(part);
		registerViewportListener(part);
		if(current != null) {	// might happen in case of error
			reportChange(OPENED, part, EMPTY_LINERANGE_LIST, current);
		}
		super.partOpened(part);
	}

	public void partActivated(IWorkbenchPart part) {
//		System.out.println("partActivated(" + part + ")");
		updateSVVisibleRegions(part);
		registerAnnotationModelListener(part);
		super.partActivated(part);
	}

	public void partBroughtToTop(IWorkbenchPart part) {
//		System.out.println("partBroughtToTop(" + part + ")");
		List<ILineRange> current = updateSVVisibleRegions(part);
		registerAnnotationModelListener(part);
		reportChange(BROUGHT_TO_TOP, part, EMPTY_LINERANGE_LIST, current);
		super.partBroughtToTop(part);
	}

	public void partDeactivated(IWorkbenchPart part) {
//		System.out.println("partDeactivated(" + part + ")");
		deregisterAnnotationModelListener(part);
		super.partDeactivated(part);
	}

	public void partClosed(IWorkbenchPart part) {
//		System.out.println("partClosed(" + part + ")");
		deregisterAnnotationModelListener(part);
		deregisterViewportListener(part);
		recentSVVisibleRegions.remove(part);
		super.partClosed(part);
	}

	private void registerAnnotationModelListener(IWorkbenchPart part) {
		if(part instanceof IEditorPart) {
			ISourceViewer sv = getSourceViewer((IEditorPart)part);
			if(sv == null) { return; }
			IAnnotationModel am = getVisualAnnotationModel(sv);
			if(am != null && !modelToPart.containsKey(am)) {
				modelToPart.put(am, (IEditorPart)part);
				am.addAnnotationModelListener(this);
			}
		}
	}

	private void deregisterAnnotationModelListener(IWorkbenchPart part) {
		if(part instanceof IEditorPart) {
			ISourceViewer sv = getSourceViewer((IEditorPart)part);
			IAnnotationModel am;
			if(sv != null && (am = getVisualAnnotationModel(sv)) != null) {
				modelToPart.remove(am);
				am.removeAnnotationModelListener(this);
			}
		}
	}

	public void modelChanged(IAnnotationModel model) {
//		System.out.println("modelChanged(" + model + ")");
		final IEditorPart part = modelToPart.get(model);
		if(part == null) { return; }
		determineChanges(FOLDING, part, getSourceViewer(part));
	}

	/**
	 * Answer the appropriate annotation model for the provided viewer
	 * corresponding to visual features (e.g., folding).  Changes to this
	 * model will be used to register interaction events corresponding to
	 * folding changes.
	 * @param viewer the viewer
	 * @return the viewer's visual annotation model
	 */
	protected IAnnotationModel getVisualAnnotationModel(ISourceViewer viewer) {
		IAnnotationModel am = null;
		if(viewer instanceof ISourceViewerExtension2 &&
				(am = ((ISourceViewerExtension2)viewer).getVisualAnnotationModel()) != null) {
			return am;
		}
		if(viewer instanceof ProjectionViewer && (am = ((ProjectionViewer)viewer).getProjectionAnnotationModel()) != null) {
			return am;
		}
		return viewer.getAnnotationModel();
	}

	private void registerViewportListener(IWorkbenchPart part) {
		ISourceViewer sv;
		if((sv = getSourceViewer(part)) != null) {
			IViewportListener pvl = createViewportListener(sv, part);
			sv.addViewportListener(pvl);
			viewportListeners.put(part, pvl);
		}		
	}

	private void deregisterViewportListener(IWorkbenchPart part) {
		ISourceViewer sv;
		if((sv = getSourceViewer(part)) != null) {
			IViewportListener vpl = viewportListeners.remove(part);
			if(vpl != null) { sv.removeViewportListener(vpl); }
		}
	}

	protected void viewportChanged(final IWorkbenchPart part, ISourceViewer viewer) {
//		System.out.println("viewportChanged(" + part + ", " + viewer + ")");
		if(part == null) { return; }
		determineChanges(SCROLLING, part, viewer);
	}


	private void determineChanges(final String changeType,
			final IWorkbenchPart part, final ISourceViewer viewer) {
		synchronized(this) {
			if(pendingViewportChanges.containsKey(viewer)) { return; }
			pendingViewportChanges.put(viewer, this);
		}
		final Display display =  part.getSite().getShell().getDisplay();
		display.asyncExec(new Runnable() {
			public void run() {
				synchronized(this) { pendingViewportChanges.remove(viewer); }
				// process anything futher in the loop first as annotation model
				// changes may cause several dispatches, etc.
				while(display.readAndDispatch()) { /*do nothing*/ }
				List<ILineRange> previous = recentSVVisibleRegions.get(part);
				List<ILineRange> visible = determineVisibleLines(part);
				if(previous == null) {
					previous = EMPTY_LINERANGE_LIST;
				} else if(visible == null || equalRanges(previous, visible)) {
					return;
				}
				reportChange(changeType, part, previous, visible);
				recentSVVisibleRegions.put(part, visible);
			}
		});		
	}

	protected int countVisibleLines(List<ILineRange> lines) {
		int count = 0;
		for(ILineRange lr : lines) {
			count += lr.getNumberOfLines();
		}
		return count;
	}
	
	protected IViewportListener createViewportListener(ISourceViewer sv, IWorkbenchPart part) {
		return new PartViewportListener(this, sv, part);
	}

	protected class PartViewportListener implements IViewportListener {
		protected IWorkbenchPart part;
		protected ISourceViewer viewer;
		protected AbstractTextEditorMonitor monitor;
		
		protected PartViewportListener(AbstractTextEditorMonitor monitor, ISourceViewer sv, IWorkbenchPart part) {
			this.monitor = monitor;
			this.part = part;
			this.viewer = sv;
		}
		
		public void viewportChanged(int verticalOffset) {
			monitor.viewportChanged(part, viewer);
		}
	}

}
