/*******************************************************************************
 * Copyright (c) 2004 - 2006 University Of British Columbia and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     University Of British Columbia - initial API and implementation
 *******************************************************************************/

package ca.ubc.cs.mylyn.monitors.parts;

import java.util.Map;
import java.util.WeakHashMap;

import org.eclipse.jface.viewers.AbstractTreeViewer;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.ITreeViewerListener;
import org.eclipse.jface.viewers.TreeExpansionEvent;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.mylyn.internal.monitor.core.util.XmlStringConverter;
import org.eclipse.mylyn.monitor.core.InteractionEvent;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.IWorkbenchPart;

/**
 * A monitor for listening to changes to a part's tree viewer.
 * The class handles the details of finding the actual part and notifying
 * when a tree viewer changes; the subclass is responsible for mapping
 * an IViewPart to the actual TreeViewer instance ({@link AbstractTreeViewerMonitor#getTreeViewer(IViewPart)}).
 * @author Brian de Alwis
 */
public abstract class AbstractTreeViewerMonitor
	extends AbstractWorkbenchPartMonitor {
	
	protected Map<TreeViewer,ITreeViewerListener> viewerListeners =
		new WeakHashMap<TreeViewer,ITreeViewerListener>();
	
	public static final String TAG_HANDLE = "handle";
	public static final String ATTR_KIND = "kind";
	public static final String TAG_TREE_EXPANDED = "treeExpanded";
	public static final String TAG_TREE_COLLAPSED = "treeCollapsed";
	
	public AbstractTreeViewerMonitor() {
		super();
		// ignore
	}

	protected abstract TreeViewer getTreeViewer(IViewPart part);

	public void partOpened(IWorkbenchPart part) {
		if(part instanceof IViewPart) {
			TreeViewer viewer = getTreeViewer((IViewPart)part);
			addViewerListener(viewer, (IViewPart)part);
		}
		super.partOpened(part);
	}
	
	public void partClosed(IWorkbenchPart part) {
		if(part instanceof IViewPart) {
			TreeViewer viewer = getTreeViewer((IViewPart)part);
			removeViewerListener(viewer, (IViewPart)part);
		}
		super.partClosed(part);
	}

	protected void addViewerListener(TreeViewer viewer, IViewPart part) {
		if(viewer == null) { return; }
		ITreeViewerListener listener = viewerListeners.get(viewer);
		if(listener == null) {
			viewerListeners.put(viewer, 
					listener = new TreeViewerListener(viewer, (IViewPart)part, this));
		}
		viewer.addTreeListener(listener);
	}

	protected void removeViewerListener(TreeViewer viewer, IViewPart part) {
		if(viewer == null) { return; }
		ITreeViewerListener listener = viewerListeners.remove(viewer);
		if(listener != null) {
			viewer.removeTreeListener(listener);
		}
	}

	protected String createChangeDelta(String eventType, 
			AbstractTreeViewer viewer, Object element, IWorkbenchPart part) {
		if(!(viewer.getContentProvider() instanceof ITreeContentProvider)) {
			return "";
		}
		ITreeContentProvider cp = (ITreeContentProvider)viewer.getContentProvider();
		StringBuffer buffer = new StringBuffer();
		buffer.append("<");
		buffer.append(eventType);
		buffer.append(">\n");
		for(Object me : cp.getChildren(element)) {
			String structureKind = getKind(me);
			buffer.append("<");
			buffer.append(TAG_HANDLE);
			buffer.append(" ");
			buffer.append(ATTR_KIND);
			buffer.append("=\"");
			buffer.append(structureKind);
			buffer.append("\">");
			buffer.append(XmlStringConverter.convertToXmlString(
					obfuscateHandle(structureKind, getHandle(me))));
			buffer.append("</");
			buffer.append(TAG_HANDLE);
			buffer.append(">\n");
		}
		buffer.append("</");
		buffer.append(eventType);
		buffer.append(">\n");
		return buffer.toString();
	}
	
	protected void treeExpanded(TreeExpansionEvent event, IViewPart part) {
		String changeDelta = createChangeDelta(TAG_TREE_EXPANDED,
				event.getTreeViewer(), event.getElement(), part);
		notify(new InteractionEvent(InteractionEvent.Kind.SELECTION,
				getKind(event.getElement()), getHandle(event.getElement()),
				part.getSite().getId(), TAG_TREE_EXPANDED, changeDelta, 0f));
	}
		
	protected void treeCollapsed(TreeExpansionEvent event, IViewPart part) {
		String changeDelta = createChangeDelta(TAG_TREE_COLLAPSED,
				event.getTreeViewer(), event.getElement(), part);
		notify(new InteractionEvent(InteractionEvent.Kind.SELECTION,
				getKind(event.getElement()), getHandle(event.getElement()),
				part.getSite().getId(), TAG_TREE_COLLAPSED, changeDelta, 0f));
	}

	protected abstract String getKind(Object me);
	protected abstract String getHandle(Object me);

	protected class TreeViewerListener implements ITreeViewerListener {
		protected AbstractTreeViewer viewer;
		protected IViewPart part;
		protected AbstractTreeViewerMonitor monitor;
		
		public TreeViewerListener(AbstractTreeViewer viewer, IViewPart part, AbstractTreeViewerMonitor monitor) {
			this.viewer = viewer;
			this.part = part;
			this.monitor = monitor;
		}
		
		public void treeExpanded(TreeExpansionEvent event) {
			monitor.treeExpanded(event, part);
		}
		
		public void treeCollapsed(TreeExpansionEvent event) {
			monitor.treeCollapsed(event, part);
		}
	}
}
