package ca.ubc.cs.mylyn.monitors.parts;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.eclipse.mylyn.internal.monitor.usage.InteractionEventObfuscator;
import org.eclipse.mylyn.internal.monitor.usage.UiUsageMonitorPlugin;
import org.eclipse.mylyn.monitor.core.InteractionEvent;
import org.eclipse.mylyn.monitor.ui.IMylarMonitorLifecycleListener;
import org.eclipse.mylyn.monitor.ui.MonitorUiPlugin;
import org.eclipse.ui.IPartListener;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPart;

/**
 * A monitor for listening to changes in workbench parts.  This class provides
 * infrastructure for listening for changes, and notifying of events.
 * Either explicitly use the {@link IMylynMonitorLifecycleListener} protocol,
 * or add to the {@link MylynUsageMonitorPlugin#addMonitoringLifecycleListener(IMylynMonitorLifecycleListener)}
 * @author Brian de Alwis
 */
public abstract class AbstractWorkbenchPartMonitor implements IPartListener, IMylarMonitorLifecycleListener {

	protected InteractionEventObfuscator obfuscator = new InteractionEventObfuscator();
	
	public AbstractWorkbenchPartMonitor() {
	}

	/**
	 * Return the part ids to be monitored.
	 */
	protected abstract String[] getPartIds();
	
	/** Default implementations: available for overriding. */

	public void partActivated(IWorkbenchPart part) {
		// ignore
	}

	public void partBroughtToTop(IWorkbenchPart part) {
		// ignore
	}

	public void partClosed(IWorkbenchPart part) {
		// ignore
	}

	public void partDeactivated(IWorkbenchPart part) {
		// ignore
	}

	public void partOpened(IWorkbenchPart part) {
		// ignore
	}

	/** Helper functions **/

	protected void notify(InteractionEvent event) {
		MonitorUiPlugin.getDefault().notifyInteractionObserved(event);
	}
	
	protected String obfuscateHandle(String kind, String handle) {
		if(!UiUsageMonitorPlugin.getDefault().isObfuscationEnabled()) {
			return handle;
		}
		return obfuscator.obfuscateHandle(kind, handle);
	}

    @SuppressWarnings("unchecked")
	protected Object invokeMethod(Object instance, String methodName,
            Class argTypes[], Object arguments[]) throws SecurityException, NoSuchMethodException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
        Class clas = instance.getClass();
        do {
    		try {
    			Method method = clas.getDeclaredMethod(methodName, argTypes);
    			method.setAccessible(true);
    			return method.invoke(instance, arguments);
    		} catch(NoSuchMethodException e) {}
    		clas = clas.getSuperclass();
        } while(clas != null);
        throw new NoSuchMethodException(methodName);
    }

    @SuppressWarnings("unchecked")
	protected Object getField(Object instance, String fieldName)
    		throws SecurityException, NoSuchFieldException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
        Class clas = instance.getClass();
        Field field = clas.getDeclaredField(fieldName);
        field.setAccessible(true);
        return field.get(instance);
    }


	/** Internal methods */
	public void startMonitoring() {
		for(String id : getPartIds()) {
			PartDispatcher.getDefault().addPartListener(id, this);
		}
		getWorkbench().getDisplay().asyncExec(new Runnable() {
			public void run() {
				for(String id : getPartIds()) {
					for(IWorkbenchPart part : PartDispatcher.getDefault().findParts(id)) {
						partOpened(part);
					}
				}
			}
		});
	}

	public void stopMonitoring() {
		for(String id : getPartIds()) {
			PartDispatcher.getDefault().removePartListener(id, this);
		}
		if(getWorkbench().isClosing()) { return; }
		getWorkbench().getDisplay().asyncExec(new Runnable() {
			public void run() {
				for(String id : getPartIds()) {
					for(IWorkbenchPart part : PartDispatcher.getDefault().findParts(id)) {
						partClosed(part);
					}
				}
			}
		});
	}

	protected IWorkbench getWorkbench() {
		return Activator.getDefault().getWorkbench();
	}

	protected String getActiveWorkbenchWindowId() {
		return getWorkbench().getActiveWorkbenchWindow().toString();
	}
}
