/*******************************************************************************
 * Copyright (c) 2004 - 2006 University Of British Columbia and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     University Of British Columbia - initial API and implementation
 *******************************************************************************/

package ca.ubc.cs.mylyn.monitors.parts;

import java.util.LinkedList;
import java.util.List;

/**
 * Compute what changed between two line ranges.  This is an abstract
 * class, intended to be suitably overridden for particular representations
 * of line ranges. 
 * @author Brian de Alwis
 * @param <I> the type representing a line range
 */
public abstract class IntervalDifferences<I> {
	protected LinkedList<I> before;
	protected LinkedList<I> after;

	protected LinkedList<I> added;
	protected LinkedList<I> removed;
	
	public IntervalDifferences(List<I> before, List<I> after) {
		this.before = new LinkedList<I>(before);
		this.after = new LinkedList<I>(after);
	}

	protected abstract int getStart(I interval);
	protected abstract int getEnd(I interval);
	protected abstract I split(I exemplar, int start, int end);
	
	protected void compute() {
		removed = new LinkedList<I>();
		added = new LinkedList<I>();
		while(!before.isEmpty() && !after.isEmpty()) {
			I b = before.getFirst();
			I a = after.getFirst();
			if(getEnd(b) < getStart(a)) {
				// b comes before a with no overlap: [--b--] [--a--]
				removed.add(before.removeFirst());
			} else if(getEnd(a) < getStart(b)) {
				// a comes before b with no overlap: [--a--] [--b--]
				added.add(after.removeFirst());
			} else if(getStart(a) < getStart(b)) {
				assert getStart(b) < getEnd(a);
				added.add(split(a, getStart(a), getStart(b) - 1));
				// if b trails past a, or if a wholly encloses b, then add remnant
				// and process on next loop iteration
				after.removeFirst();
				before.removeFirst();
				if(getEnd(a) < getEnd(b)) {
					before.addFirst(split(b, getEnd(a) + 1, getEnd(b)));
				} else if(getEnd(a) > getEnd(b)) {
					after.addFirst(split(a, getEnd(b) + 1, getEnd(a)));
				}
			} else if(getStart(b) < getStart(a)) {
				assert getStart(a) < getEnd(b);
				removed.add(split(b, getStart(b), getStart(a) - 1));
				// if a trails past b, or if b wholly encloses a, then add remnant
				// and process on next loop iteration
				after.removeFirst();
				before.removeFirst();
				if(getEnd(b) < getEnd(a)) {
					after.addFirst(split(a, getEnd(b) + 1, getEnd(a)));
				} else if(getEnd(b) > getEnd(a)) {
					before.addFirst(split(b, getEnd(a) + 1, getEnd(b)));
				}
			} else {
				if(getStart(a) != getStart(b)) {
					throw new AssertionError("algorthmic error: getStart(a) != getStart(b)");
				}
				after.removeFirst();	
				before.removeFirst();
				if(getEnd(a) < getEnd(b)) {
					//	a is completely overlapped by b
					before.addFirst(split(b, getEnd(a) + 1, getEnd(b)));
				} else if(getEnd(b) < getEnd(a)) {
					//	b is completely overlapped by a
					after.addFirst(split(a, getEnd(b) + 1, getEnd(a)));					
				} else { // getEnd(a) == getEnd(b)
					// do nothing since complete overlap: remove both and move on
				}
			}
		}
		if(!before.isEmpty()) {
			removed.addAll(before);
		} else if(!after.isEmpty()) {
			added.addAll(after);
		}
	}

	public LinkedList<I> getAdded() {
		if(added == null) { compute(); }
		return added;
	}

	public LinkedList<I> getRemoved() {
		if(removed == null) { compute(); }
		return removed;
	}
}
