/*******************************************************************************
 * Copyright (c) 2004 - 2006 University Of British Columbia and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     University Of British Columbia - initial API and implementation
 *******************************************************************************/

package ca.ubc.cs.mylyn.monitors.parts;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.ListenerList;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.IPageChangeProvider;
import org.eclipse.mylyn.monitor.ui.MonitorUiPlugin;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.ScrollBar;
import org.eclipse.swt.widgets.Scrollable;
import org.eclipse.swt.widgets.Widget;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IPartListener;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.IViewReference;
import org.eclipse.ui.IViewSite;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchPartReference;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.internal.PartPane;
import org.eclipse.ui.internal.PartSite;
import org.eclipse.ui.internal.WorkbenchPartReference;
import org.eclipse.ui.part.MultiPageEditorPart;

/**
 * A utility class for monitoring and dispatching to listeners wanting notification
 * of updates to particular classes of part.  Two types of updates are provided:
 * part lifecycle notifications (as per {@link IPartListener}), and scrolling notification
 * (as per {@link IPartScrollingListener}).  Listeners are responsible for ensuring
 * they are removed.
 * @author bsd
 * @see org.eclipse.ui.IPartListener
 * @see ca.ubc.cs.mylyn.monitors.parts.IPartScrollingListener
 */
@SuppressWarnings("restriction")
public class PartDispatcher {
	protected static PartDispatcher singleton;
	
	public static PartDispatcher getDefault() {
		if(singleton == null) {
			singleton = new PartDispatcher();
		}
		return singleton;
	}

	public static void disposeDefault() {
		if(singleton != null) { singleton.dispose(); }
	}
	
	/**
	 * Listens as a global SWT event filter, allowing us to be notified of scrolling
	 * events. 
	 */
	private Listener displayScrollingFilter;

	/**
	 * Listens and dispatches notifications of part changes.
	 */
	private IPartListener partListener;

	/**
	 * The lists of listeners interesting in being notified of widget scrolling or
	 * part lifecycle events for particular parts.  Keyed by the part Ids. 
	 */
	private Map<String,ListenerList> scrollingListeners =
		new HashMap<String,ListenerList>();
	private Map<String,ListenerList> partListeners =
		new HashMap<String,ListenerList>();
	
	/**
	 * Lookup cache for controls to their appropriate part.  Note that
	 * initializePartMapping() assigns this as a synchronized map.
	 */
	private Map<Control,IWorkbenchPart> parts;
	
	/**
	 * See {@link PartDispatcher#handleScrollingEvent(Event)}.
	 */
	private Map<ScrollBar,Runnable> pendingScrollingNotifications;
	
	public synchronized void addScrollingListener(String partId, IPartScrollingListener listener) {
		ListenerList partScrollingListeners = scrollingListeners.get(partId);
		if(partScrollingListeners == null) {
			scrollingListeners.put(partId, partScrollingListeners = new ListenerList()); 
		}
		partScrollingListeners.add(listener);
		for(IWorkbenchPart part : findParts(partId)) {
			recordPartPane(part);
		}
		ensureRegistered();
	}

	public synchronized void removeScrollingListener(String partId, IPartScrollingListener listener) {
		ListenerList partScrollingListeners = scrollingListeners.get(partId);
		if(partScrollingListeners != null) {
			partScrollingListeners.remove(listener);
			if(partScrollingListeners.isEmpty()) {
				scrollingListeners.remove(partId); 
			}
		}
		checkDispose();
	}
	
	public synchronized void addPartListener(String id, IPartListener listener) {
		ListenerList listeners = partListeners.get(id);
		if(listeners == null) {
			partListeners.put(id, listeners = new ListenerList());
		}
		listeners.add(listener);
		ensureRegistered();
	}

	public synchronized void removePartListener(String id, IPartListener listener) {
		ListenerList listeners = partListeners.get(id);
		if(listeners != null) {
			listeners.remove(listener);
			if(listeners.isEmpty()) {
				partListeners.remove(id);
			}
		}
		checkDispose();
	}

	private void checkDispose() {
		if(scrollingListeners.isEmpty() && partListeners.isEmpty()) {
			dispose();
		}		
	}

	private void ensureRegistered() {
		// implicit test of whether registered
		if(displayScrollingFilter == null && !scrollingListeners.isEmpty()) {
			initializePartMapping();
			initializeDisplayFiltering();
		}
		
		if(partListener == null && !partListeners.isEmpty()) {
			initializePartMapping();
		}
	}
	
	private void initializePartMapping() {
		if(parts != null) { return; }	// already initialized
		parts = Collections.synchronizedMap(new WeakHashMap<Control,IWorkbenchPart>());
		pendingScrollingNotifications = Collections.synchronizedMap(new WeakHashMap<ScrollBar,Runnable>());
		partListener = createPartListener();
		MonitorUiPlugin.getDefault().addWindowPartListener(partListener);
		// Do we want to iterate over all current parts and do partListener.partOpened(p)?
	}

	private void initializeDisplayFiltering() {
		if(displayScrollingFilter != null) { return; }	// implicit test
		displayScrollingFilter = createScrollingListener();
		getWorkbench().getDisplay().asyncExec(new Runnable() {
			public void run() {
				getWorkbench().getDisplay().addFilter(SWT.Selection, displayScrollingFilter);
			}
		});
	}
	
	public void dispose() {
		if(partListener != null) {
			MonitorUiPlugin.getDefault().removeWindowPartListener(partListener);
			partListener = null;
		}
		if(displayScrollingFilter != null && !getWorkbench().getDisplay().isDisposed()) {
			getWorkbench().getDisplay().syncExec(new Runnable() {
				public void run() {
					if(displayScrollingFilter != null) {
						getWorkbench().getDisplay().removeFilter(SWT.Selection, displayScrollingFilter);
						displayScrollingFilter = null;
					}
				}
			});
		}
		parts = null;
		if(this == singleton) { singleton = null; }
	}
	
	private IWorkbench getWorkbench() {
		return Activator.getDefault().getWorkbench();
	}

	protected IPartListener createPartListener() {
		return new IPartListener() {
			public void partOpened(IWorkbenchPart part) {
				reportPartOpened(part);
				recordPartPane(part);
			}
			
			public void partClosed(IWorkbenchPart part) {
				reportPartClosed(part);
				removePartPane(part);
			}
			
			public void partActivated(IWorkbenchPart part) {
				reportPartActivated(part);
			}
			
			public void partBroughtToTop(IWorkbenchPart part) {
				reportPartBroughtToTop(part);
			}
			
			public void partDeactivated(IWorkbenchPart part) {
				reportPartDeactivated(part);
			}
		};
	}

	protected Listener createScrollingListener() {
		return new Listener() {
			public void handleEvent(Event event) {
				if ((event.type != SWT.Selection && event.type != SWT.DefaultSelection)
						|| !(event.widget instanceof ScrollBar)) {
					return;
				}
				handleScrollingEvent(event);
			}
		};
	}
	
	protected void handleScrollingEvent(final Event event) {
		// Note: it appears that with 3.3M3 (at least), scrolling events don't happen
		// before the actual scroll.  Previously we were able to lodge an async-exec to
		// notify of the post-scroll, providing listeners the ability to get a pre- and
		// post-scroll.  Now we just notify post-scroll only.
		if(event.count > 0) { return; }
		// We register the notifier for this ScrollBar; the notifier checks that it
		// is the latest notifier before executing, so as to ensure we minimize the
		// number of intermediate 
		final ScrollBar scrollBar = (ScrollBar)event.widget;
		Runnable notifier = new Runnable() {
			public void run() {
				// skip all notifiers except for the last one.
				if(pendingScrollingNotifications.get(scrollBar) != this) { return; }
				final Scrollable scrollable = scrollBar.getParent();
//				System.out.println("Scrolling: " + scrollable);
				final IWorkbenchPart part = findCorrespondingPart(scrollable);
				final IWorkbenchPart embeddedPart = getEmbeddedPart(part);
//				System.out.println("Scrolling: part=" + part);
				if(part != null) {
					reportPartScroll(part.getSite().getId(), part, scrollable, scrollBar);
				}
				if(embeddedPart != null) {
					reportPartScroll(embeddedPart.getSite().getId(), embeddedPart, scrollable, scrollBar);
				}
			}};
			pendingScrollingNotifications.put(scrollBar, notifier);
			getWorkbench().getDisplay().asyncExec(notifier);

	}
		
	private void reportPartScroll(String partId, IWorkbenchPart part, Scrollable scrollable, ScrollBar scrollBar) {
		ListenerList listeners = scrollingListeners.get(partId);
		if(listeners != null) {
			for(Object listener : listeners.getListeners()) {
				((IPartScrollingListener)listener).partScrolled(partId, part, scrollable, scrollBar);
			}
		}
	}
	
	private IWorkbenchPart findCorrespondingPart(Widget widget) {
		Widget last = null;
		IWorkbenchPart part = null;
		// List<Widget> path = new ArrayList<Widget>();
		do {
			if(parts.containsKey(widget)) {
				part = parts.get(widget);
				// System.out.println("event widget is part: " + part);
				// System.out.println("  chain: " + path);
				return part;
			}
			// path.add(widget);
			last = widget;
			if(widget instanceof Control) {
				widget = ((Control)widget).getParent();
			} else if(widget instanceof ScrollBar) {
				widget = ((ScrollBar)widget).getParent();
			}
		} while(widget != null && widget != last);
		return null;
	}

	private IEditorReference findEditorReference(IEditorSite site) {
		for(IEditorReference er : site.getPage().getEditorReferences()) {
			if(er.getId().equals(site.getId())) { return er; }
		}
		return null;
	}

	// Public only for testing
	public Listener getDisplayScrollingFilter() {
		return displayScrollingFilter;
	}

	protected IWorkbenchPart getEmbeddedPart(IWorkbenchPart part) {
		if(part instanceof MultiPageEditorPart && part instanceof IPageChangeProvider) {
			Object page = ((IPageChangeProvider)part).getSelectedPage();
			if(page instanceof IEditorPart) {
				return (IEditorPart)page;
			}
		}
		return null;
	}

	private void reportPartDeactivated(IWorkbenchPart part) {
		ListenerList list = partListeners.get(part.getSite().getId());
		if(list == null) { return; }
		for(Object listener : list.getListeners()) {
			((IPartListener)listener).partDeactivated(part);
		}
	}

	private void reportPartActivated(IWorkbenchPart part) {
		ListenerList list = partListeners.get(part.getSite().getId());
		if(list == null) { return; }
		for(Object listener : list.getListeners()) {
			((IPartListener)listener).partActivated(part);
		}
	}

	private void reportPartBroughtToTop(IWorkbenchPart part) {
		ListenerList list = partListeners.get(part.getSite().getId());
		if(list == null) { return; }
		for(Object listener : list.getListeners()) {
			((IPartListener)listener).partBroughtToTop(part);
		}
	}

	private void reportPartOpened(IWorkbenchPart part) {
		ListenerList list = partListeners.get(part.getSite().getId());
		if(list == null) { return; }
		for(Object listener : list.getListeners()) {
			((IPartListener)listener).partOpened(part);
		}
	}

	private void reportPartClosed(IWorkbenchPart part) {
		ListenerList list = partListeners.get(part.getSite().getId());
		if(list == null) { return; }
		for(Object listener : list.getListeners()) {
			((IPartListener)listener).partClosed(part);
		}
	}

	public Collection<IWorkbenchPart> findParts(String partId) {
		Set<IWorkbenchPart> parts = new HashSet<IWorkbenchPart>();
		for (IWorkbenchWindow window : getWorkbench().getWorkbenchWindows()) {
			// should we be iterating over all pages?
			IWorkbenchPage page = window.getActivePage(); 
			if(page == null) { continue; }
			for(IViewReference vr : page.getViewReferences()) {
				if(partId.equals(vr.getId())) {
					IViewPart view = vr.getView(false);
					if(view != null) { parts.add(view); }
				}
			}
			for(IEditorReference er : page.getEditorReferences()) {
				if(partId.equals(er.getId())) {
					IEditorPart editor = er.getEditor(false);
					if(editor != null) { parts.add(editor); }
				}
			}
		}

		return parts;
	}

	protected void recordPartPane(IWorkbenchPart part) {
		PartPane pane = null;
		// This first if clause should cover everything
		if(part.getSite() instanceof PartSite) {
			IWorkbenchPartReference pr = ((PartSite)part.getSite()).getPartReference();
			if(pr == null || !(pr instanceof WorkbenchPartReference)) { return; }
			pane = ((WorkbenchPartReference)pr).getPane();
		} else  if(part instanceof IViewPart) {
			IViewSite site = (IViewSite)part.getSite();
			IViewReference vr = site.getPage().findViewReference(site.getId(), site.getSecondaryId());
			if(vr == null || !(vr instanceof WorkbenchPartReference)) { return; }
			pane = ((WorkbenchPartReference)vr).getPane();
		} else if(part instanceof IEditorPart) {
			IEditorSite site = (IEditorSite)part.getSite();
			IEditorReference er = findEditorReference(site);
			if(er == null || !(er instanceof WorkbenchPartReference)) { return; }
			pane = ((WorkbenchPartReference)er).getPane();			
		} else {
			Activator.getDefault().getLog().log(new Status(IStatus.WARNING,
					Activator.PLUGIN_ID, -1, getClass() + ": unknown part type: " + part.getClass(), null));
		}
		if(pane != null && pane.getControl() != null) {
//					System.out.println("  Associating part " + e + " to " + pane.getControl());
			synchronized(parts) {
				parts.put(pane.getControl(), part);
			}
		}
	}

	protected void removePartPane(IWorkbenchPart part) {
		synchronized(parts) {
			if(!parts.containsValue(part)) { return; }
			for(Iterator<Control> it = parts.keySet().iterator(); it.hasNext();) {
				Control c = it.next();
				if(parts.get(c) == part) {
					it.remove();
				}
			}
		}
	}

}
