package ca.ubc.cs.mylyn.monitors.parts.analysis;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.eclipse.jface.text.source.ILineRange;
import org.eclipse.jface.text.source.LineRange;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import ca.ubc.cs.mylyn.monitors.parts.AbstractTextEditorMonitor;
import ca.ubc.cs.mylyn.monitors.parts.jdt.JavaEditorViewportMonitor;

/**
 * Describe the changes to an editor's viewable region, in terms of new lines
 * added, lines removes, and the total lines visible, as well as describing the
 * model elements (or handles) newly visible as a result of the change
 * described. Handles are described in two parts: a <em>kind</em> describing
 * the type of handle, and the identifier, which is some <em>kind</em>-specific
 * string.
 * 
 * <p>
 * <B>NOTE:</B> We recommend ignoring viewports from
 * {@link AbstractTextEditorMonitor#OPENED OPENED} events when seeking
 * information about the visible lines.
 * {@link AbstractTextEditorMonitor#OPENED OPENED} appears to be sent the first
 * time a file is opened (i.e., when a new editor was necessary to display the
 * file's contents). But such events report the lines that <em>would</em> be
 * visible if the entirety of the file could be displayed, which is usually the entirety of
 * the file (with only the imports and copyright folded), rather than what <EM>was
 * actually</EM> displayed in the viewport. These
 * {@link AbstractTextEditorMonitor#OPENED OPENED} events seem to always be
 * followed by a {@link AbstractTextEditorMonitor#BROUGHT_TO_TOP BROUGHT_TO_TOP}
 * which do correctly list what is <em>actually</em> displayed.
 * </p>
 * 
 * TODO: generalize to support more than just &lt;newlyVisibleElements&gt;
 * TODO: expose add*() methods, and use for *generating* the XML too?
 * 
 * @author Brian de Alwis
 */
public class ViewportPayload {
	protected String type;
	protected List<ILineRange> added = new ArrayList<ILineRange>();
	protected List<ILineRange> removed = new ArrayList<ILineRange>();
	protected List<ILineRange> visible = new ArrayList<ILineRange>();
	protected Map<String[],ILineRange> handles = new HashMap<String[],ILineRange>();
	
	/**
	 * Parse from the provided XML description.
	 * @param viewportDescription
	 * @return the parsed payload
	 */
	public static ViewportPayload load(String viewportDescription) {
		return new ViewportPayload().loadFromXML(viewportDescription);
	}
	
	/**
	 * Return the number of visible lines.
	 * @return the number of visible lines
	 */
	public int getNumberVisibleLines() {
		int nv = 0;
		for(ILineRange lr : visible) {
			nv += lr.getNumberOfLines();
		}
		return nv;
	}
	
	/**
	 * Return the type of the change (e.g., <em>scrolling</em>,
	 * <em>brought-to-top</em>, <em>folding</em>).
	 * @return
	 */
	public String getType() {
		return type;
	}
	
	public List<ILineRange> getRemoved() {
		return removed;
	}

	public List<ILineRange> getVisible() {
		return visible;
	}

	public Map<String[], ILineRange> getHandles() {
		return handles;
	}

	public List<ILineRange> getAdded() {
		return added;
	}

	protected ViewportPayload loadFromXML(String viewportDescription) {
		PayloadHandler handler = new PayloadHandler();
        SAXParserFactory factory = SAXParserFactory.newInstance();
		try {
	        SAXParser saxParser = factory.newSAXParser();	        
	        saxParser.parse(new StringInputStream(viewportDescription), handler);
		} catch (ParserConfigurationException e) {
			throw new RuntimeException("Parser configuration exception", e);
		} catch (SAXException e) {
			throw new RuntimeException("SAX parsing exception", e);
		} catch (IOException e) {
			throw new RuntimeException("IO exception", e);
		}
		return this;
	}

	public static List<ILineRange> parseLineRange(String value) {
		List<ILineRange> range = new ArrayList<ILineRange>();
		for(String r : value.split("[ \t\n]")) {
			if(r.length() > 0) {
				range.add(parseSingleLineRange(r));
			}
		}
		return range;
	}

	public static ILineRange parseSingleLineRange(String r) {
		int colon = r.indexOf(':');
		if(colon < 0) {
			int line = Integer.parseInt(r.trim());
			return new LineRange(line, 1);
		} else {
			int start = Integer.parseInt(r.substring(0, colon).trim());
			int stop = Integer.parseInt(r.substring(colon+1).trim()); 
			return new LineRange(start, stop - start + 1);
		}
	}

	protected void setType(String _type) {
		type = _type;
	}
	
	protected void addAdded(List<ILineRange> lr) {
		added.addAll(lr);
	}

	protected void addVisible(List<ILineRange> lr) {
		visible.addAll(lr);
	}

	protected void addRemoved(List<ILineRange> lr) {
		removed.addAll(lr);
	}

	protected void addHandle(String handleType, String trim, String handleLines) {
		handles.put(new String[] { handleType, trim}, parseSingleLineRange(handleLines));
	}


	/**
	 * Read in a viewport description as generated by the AbstractTextEditorMonitor
	 * and other monitors of its ilk (i.e., description of what was visible within
	 * a JavaEditor).  Viewport descriptions describe the actual lines visible
	 * (and their delta from the previous report: what lines were added and what
	 * lines were removed), as well as the actual elements visible.
	 * @author bsd
	 */
	public class PayloadHandler extends DefaultHandler {
		protected StringBuffer characters;		
		protected String handleType;
		protected String handleLines;
		
		protected PayloadHandler() {
		}
		
		@Override
		public void startElement(String uri, String localName, String qName, Attributes attrs) throws SAXException {
			// subclasses should instead override startElement(String,Attributes),
			// unless you really know what you're doing
			startElement(uri, localName == null || localName.length() == 0 ? qName : localName, 
					attrs);
			characters = new StringBuffer();
		}
		
		protected void startElement(String uri, String elementName, Attributes attrs) {
			if(elementName.equals(JavaEditorViewportMonitor.TAG_NEWLY_VISIBLE_ELEMENTS)) {
				// do nothing: we'll pick up on TAG_HANDLE
			} else if(elementName.equals(JavaEditorViewportMonitor.FOLDING) 
					|| elementName.equals(JavaEditorViewportMonitor.SCROLLING)
					|| elementName.equals(JavaEditorViewportMonitor.OPENED)
					|| elementName.equals(JavaEditorViewportMonitor.BROUGHT_TO_TOP)) {
				setType(elementName);
				addVisible(parseLineRange(getValue(attrs, uri, JavaEditorViewportMonitor.ATTR_VISIBLE)));
				addAdded(parseLineRange(getValue(attrs, uri, JavaEditorViewportMonitor.ATTR_ADDED)));
				addRemoved(parseLineRange(getValue(attrs, uri, JavaEditorViewportMonitor.ATTR_REMOVED)));
			} else if(elementName.equals(JavaEditorViewportMonitor.TAG_HANDLE)) {
	        	handleType = getValue(attrs, uri, JavaEditorViewportMonitor.ATTR_KIND);
	        	handleLines = getValue(attrs, uri, JavaEditorViewportMonitor.ATTR_LINES);
			} else {
				throw new RuntimeException("unknown tag: " + elementName);
			}			
		}

		@Override
		public void characters(char[] ch, int offset, int length) throws SAXException {
			characters.append(ch, offset, length);
		}

		@Override
		public void endElement(String uri, String localName, String qName) throws SAXException {
			endElement(uri,  localName == null || localName.length() == 0 ? qName : localName);
		}
		
		protected void endElement(String url, String elementName) {
			if(elementName.equals(JavaEditorViewportMonitor.TAG_NEWLY_VISIBLE_ELEMENTS)) {
				// do nothing: we'll pick up on TAG_HANDLE
			} else if(elementName.equals(JavaEditorViewportMonitor.FOLDING) 
					|| elementName.equals(JavaEditorViewportMonitor.SCROLLING)
					|| elementName.equals(JavaEditorViewportMonitor.OPENED)
					|| elementName.equals(JavaEditorViewportMonitor.BROUGHT_TO_TOP)) {
				// do nothing: we'll pick it up on TAG_NEWLY_VISIBLE_ELEMENTS
			} else if(elementName.equals(JavaEditorViewportMonitor.TAG_HANDLE)) {
				addHandle(handleType, characters.toString().trim(), handleLines);
				handleType = handleLines = null;
			} else {
				throw new RuntimeException("unknown tag: " + elementName);
			}
		}

		protected String getValue(Attributes a, String uri, String name) {
	        String value = a.getValue(uri, name);
	        if (value == null) {
	            value = a.getValue(name);
	        }
	        return value != null ? value.trim() : null;
	    }
	}
}
