package ca.ubc.cs.mylyn.monitors.parts.jdt;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.ISourceReference;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.internal.ui.actions.SelectionConverter;
import org.eclipse.jdt.internal.ui.javaeditor.JavaEditor;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.text.TextSelection;
import org.eclipse.jface.text.source.ILineRange;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.text.source.LineRange;
import org.eclipse.ui.IWorkbenchPart;

import ca.ubc.cs.mylyn.monitors.parts.AbstractModelBasedTextEditorMonitor;

/**
 * <P>The <TT>JavaEditorViewportMonitor</TT> causes editor changes to be
 *  recorded as <TT>selection</TT> events, with the <TT>delta</TT> 
 *  containing an XML fragment such as:</P>
 *  <pre>
 *  &lt;interactionEvent&gt;
 * 		&lt;kind&gt;selection&lt;/kind&gt;
 * 		<EM>&lt;!-- other tags elided --&gt;</EM>
 * 		&lt;navigation&gt;scrolling&lt;/navigation&gt;
 *		&lt;delta&gt;
 * 			&lt;scrolling removed=&quot;23 &quot; added=&quot;64 &quot; visible=&quot;24:26 31:64 &quot;&gt;
 *         		&lt;newlyVisibleElements&gt;
 * 					&lt;handle kind=&quot;java&quot; lines=&quot;41:150&quot;&gt;
 * 						=jEdit-4.1pre6/&amp;lt;org.gjt.sp.jedit.options{LoadSaveOptionPane.java[LoadSaveOptionPane~_init
 * 					&lt;/handle&gt;
 * 				&lt;/newlyVisibleElements&gt;
 * 			&lt;/scrolling&gt;
 * 		&lt;/delta&gt;
 * 	&lt;/interactionEvent&gt;
 * </pre>
 * <P>There are four types of delta, corresponding to the 
 * <TT>navigation</TT> type:
 * <DL>
 * <DT><TT>&lt;opened&gt;</TT></DT>
 *     <DD>this part (an editor) was newly opened</DD>
 * <DT><TT>&lt;brought-to-top&gt;</TT></DT>
 *     <DD>the part was just brought to the top</DD>
 * <DT><TT>&lt;folding&gt;</TT></DT>
 *     <DD>there was a folding change, resulting in either
 *     	an expansion or contraction in content</DD>
 * <DT><TT>&lt;scrolling&gt;</TT></DT>
 *     <DD>the part was scrolled</DD>
 * </DL>
 * <P>
 * All delta types have three attributes. 
 * The <TT>added</TT> and <TT>removed</TT> attributes report the change
 * since the previously reported delta.  The <TT>visible</TT> attribute
 * summarizes what is currently visible.  Each attribute reports a
 * set of space-delimited numbers representing line ranges.</P>
 * 
 * <P>The delta also reports the handles of the domain-specific elements
 * that were <EM>added</EM> as a result of the change (where available).
 * This is a delta from the previously reported changes, and is listed in the\
 * <TT>&lt;newlyVisibleElements&gt;</TT>...<TT>&lt;/newlyVisibleElements&gt;</TT>
 * tags. Each handle represents some domain-level element (the domain is
 * contained in the <TT>kind</TT> attribute); the handle's <TT>lines</TT>
 * attribute provides the line range containing the elements definition.</P>
 * 
 * <P>The example above reports that a part was scrolled, with line
 * 23 disappearing, line 64 newly appearing, and the net result being 
 * lines 24 through 26 and 31 through 64 are currently visible.  The
 * scrolling revealed more of an element (from its <TT>lines</TT>
 * attribute, we can determine that part of the definition was previously
 * visible).  </P>
 * @author Brian de Alwis
 */
@SuppressWarnings("restriction")
public class JavaEditorViewportMonitor extends AbstractModelBasedTextEditorMonitor<IJavaElement> {

	protected String[] getPartIds() {
		return new String[] {
				JavaUI.ID_CF_EDITOR, JavaUI.ID_CU_EDITOR
		};
	}
	
	public JavaEditorViewportMonitor() {
	}

	@Override
	protected ISourceViewer getSourceViewer(IWorkbenchPart part) {
		if(part instanceof JavaEditor) {
			return ((JavaEditor)part).getViewer();
		}
		return super.getSourceViewer(part);
	}
	@Override
	protected Collection<IJavaElement> determineCorrespondingElements(
			IWorkbenchPart part, IDocument document , 
			List<ILineRange> changed) {
		JavaEditor javaEditor = (JavaEditor)part;
		// Note that all line numbers are 0-relative *until they are made
		// palatable for human display*
		IJavaElement input= SelectionConverter.getInput(javaEditor);
		Set<IJavaElement> results = new HashSet<IJavaElement>();
		for(ILineRange lr : changed) {
			for(int count = 0; count < lr.getNumberOfLines(); count++) {
				try {
					int lineNumber = count + lr.getStartLine();
					int startOffset = document.getLineOffset(lineNumber);
					int length = document.getLineLength(lineNumber) - 1;
					// for getElementAtOffset() to work properly, must remove
					// starting and trailing whitespace
					while(length > 0 && Character.isWhitespace(document.getChar(startOffset))) {
						startOffset++;
						length--;
					}
					while(length > 0 && Character.isWhitespace(document.getChar(startOffset + length - 1))) {
						length--;
					}
					if(length == 0) { continue; }
					ITextSelection selection = new TextSelection(document, startOffset, length);
					IJavaElement elmt = SelectionConverter.getElementAtOffset(input, selection);
					if(elmt != null && elmt instanceof IMember) {
						results.add(elmt);
					}
				} catch (BadLocationException e) { 
					/* do nothing */
				} catch(JavaModelException e) {
					/* do nothing */
				}
			}
		}
		return results;
	}

	@Override
	protected String getStructureKind(IWorkbenchPart part) {
		return "java";	// JavaStructureBridge.CONTENT_TYPE
	}

	@Override
	protected String getStructureHandle(IWorkbenchPart part) {
		return SelectionConverter.getInput((JavaEditor)part).getHandleIdentifier();
	}

	@Override
	protected ILineRange getLineRange(IDocument document, IJavaElement je) {
		if(je instanceof ISourceReference) {
			try {
				ISourceReference sr = (ISourceReference)je;
				int startLine = document.getLineOfOffset(sr.getSourceRange().getOffset());
				int stopLine = document.getLineOfOffset(sr.getSourceRange().getOffset() + sr.getSourceRange().getLength() - 1);
				return new LineRange(startLine, stopLine - startLine + 1);
			} catch(JavaModelException e) {
				/* do nothing */
			} catch(BadLocationException e) {
				/* do nothing */
			}
		}
		return null;
	}

	@Override
	protected String getKind(IJavaElement je) {
		return "java";
	}

	@Override
	protected String getHandle(IJavaElement je) {
		return je.getHandleIdentifier();
	}
}
